/* -*- linux-c -*- (Autoset linux indent style when using emacs) */

/*
 * Copyright 1999 by Digi International (www.digi.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the 
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *----------------------------------------------------------------------
 *  Module File:  dgdm_isdn.c
 *
 *  Purpose:
 *
 *      
 *
 *
 *  Revision History:
 *    $Log: dgdm_isdn.c,v $
 *    Revision 1.6  2001/04/05 15:46:34  randall
 *    - Updates to support Linux 2.4
 *
 *    Revision 1.5  2000/03/23 20:45:16  randall
 *    - Implement the booting of boards in parallel using lib/triboot.
 *
 *    Revision 1.4  1999/12/11 00:20:39  randall
 *    - Complete code cleanup:
 *      - All functions prototyped.
 *      - Global includes, #defines, and structs all moved to the new
 *        dgdm_driver.h
 *      - eqm.c, audiotap.c and mgmt.c renamed with "dgdm_" prefixes
 *      - Global functions renamed to have "dgdm_" prefixes.
 *      - Unused functions removed.
 *      - #if 0 code also removed.
 *      - Corrected all instances of #include "file.c"
 *      - Makefile modified to stop including libc.a in final linking
 *      - Makefile modified to link in the tracing only when tracing
 *        enabled in the Makefile.
 *    - Tracebuffer changed to use dynamically allocated memory, thus
 *      allowing arbitrary sizes without recompilation.
 *    - Fastbaud functionality removed from the ioctls in dgdm_tty.c
 *      as per Functional Specification changes.
 *
 *    Revision 1.3  1999/09/03 14:54:44  dickl
 *    Added the GNU Public License to pretty much everything
 *    Minor Makefile changes
 *
 *    Revision 1.2  1999/08/12 18:58:34  randall
 *     - Change to using the triboot library.  This involved changing the BOARD
 *       typedef to be BOARD_T to prevent a conflict with the BOARD typedef in
 *       tribl.h
 *     - Most of the remaining changes involved removing now obsolete code from
 *       dgdm_driver.c that was used in the old boot proceedure.
 *     - No longer identify hercules boards with an artificial revision number
 *     - Removed support for pre-production HLC boards such as the Gal9.
 *
 *    Revision 1.1  1999/06/23 22:08:25  randall
 *    The great renaming from digi_dxb to dgdm as per Rick Richardson's request.
 *
 *    Revision 1.4  1999/05/19 17:17:48  dickl
 *    Correct comments that were generating compiler warnings
 *
 *    Revision 1.3  1999/05/12 13:59:00  dickl
 *    Minor reformatting
 *
 *
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 *  Local Include Files
 *----------------------------------------------------------------------
 */
#define __NO_VERSION__
#include "dgdm_driver.h"
#include "dgdm_isdn.h"            /* User data structure definitions. */

/*
 *----------------------------------------------------------------------
 *  External Variables
 *----------------------------------------------------------------------
 */
extern int		dgdm_debug;
extern int		dgdm_tty;
extern int		dgdm_isdn;
extern BOARD_T		*dgdm_Board[MAXBOARDS];
extern uchar		dgdm_NumBoards;

/*
 *----------------------------------------------------------------------
 *  MACROS
 *----------------------------------------------------------------------
 */
#define INTERNAL_COUNTERS   (sizeof(ushort)*2)
#define RX_BUFFER_OVERHEAD  1024

/*
 * Place the computed amount of free space in bytes.
 */
#define RX_BUFFER_FREE(channel, bytes)                          \
        {                                                       \
            bytes = channel->rx_buffer_size - channel->rx_buffer_bytes; \
        }

/*
 * Modify the offset based upon the internal control and data length
 * count representation.
 */
#define RX_HDR_INCR(channel, offset)                            \
        {                                                       \
            offset += ( INTERNAL_COUNTERS );                    \
            if(offset >= channel->rx_buffer_size)              \
                offset = 0;                                     \
        }

/*
 * Modify the offset by count and perform any wrap calculations and
 * alignment adjustments.
 */
#define RX_INCR(channel, offset, count)                         \
        {                                                       \
            offset += count;                                    \
            offset += 3;    /* adjust to next 4 byte address. */\
            offset &= ~3;   /* place on 4 byte boundary. */     \
            if(offset >= channel->rx_buffer_size)              \
                offset = 0;                                     \
        }

/*
 * Return the number of bytes until the end of the buffer OR count bytes.
 */
#define RX_MAX_OUT(channel, max_count, copy_bytes)              \
        {                                                       \
            if(channel->rx_in > channel->rx_out)               \
                copy_bytes = channel->rx_in - channel->rx_out;  \
            else                                                \
                copy_bytes = channel->rx_buffer_size - channel->rx_out; \
                                                                \
            if(copy_bytes > max_count)                         \
                copy_bytes = max_count;                         \
                                                                \
            channel->rx_buffer_bytes -= ((copy_bytes+3)&~3);    \
        }

/*
 * Return the number of bytes until the end of the buffer OR count bytes.
 */
#define RX_MAX_IN(channel, offset, max_count, copy_bytes)       \
        {                                                       \
            if(channel->rx_out > offset)                       \
                copy_bytes = channel->rx_out - offset;          \
            else                                                \
                copy_bytes = channel->rx_buffer_size - offset;  \
                                                                \
            if(copy_bytes > max_count)                         \
                copy_bytes = max_count;                         \
                                                                \
            channel->rx_buffer_bytes += ((copy_bytes+3)&~3);    \
        }


/*
 *----------------------------------------------------------------------
 *  External/Local Forward Reference Functions
 *----------------------------------------------------------------------
 */

STATIC int		dgdm_isdn_read_channel(BOARD_T *board,int device_index,
                                               char *data, int bytes,
                                               loff_t *ppos);
STATIC int		dgdm_isdn_write_channel(BOARD_T *board,
                                                int device_index,
                                                const char *data, int bytes,
                                                loff_t *ppos);
STATIC int		dgdm_isdn_open_channel(BOARD_T *board,DXBSTRM *channel,
                                               DXB_STRM_MESSAGE *msg,
                                               const char *data, int bytes);
STATIC int		dgdm_isdn_close_channel(BOARD_T *board,
                                                int device_index);
STATIC int		dgdm_isdn_write_data(BOARD_T *board, DXBSTRM *channel,
                                             DXB_STRM_MESSAGE *msg);
STATIC int		dgdm_isdn_push(BOARD_T *board, DXBSTRM *channel,
                                       DXB_STRM_MESSAGE *msg,const char *data);
STATIC int		dgdm_isdn_pop(BOARD_T *board, DXBSTRM *channel,
                                      DXB_STRM_MESSAGE *msg, const char *data);
STATIC int		dgdm_isdn_link(BOARD_T *board, DXBSTRM *channel,
                                       DXB_STRM_MESSAGE *msg,const char *data);
STATIC int		dgdm_isdn_unlink(BOARD_T *board, DXBSTRM *channel,
                                         DXB_STRM_MESSAGE *msg,
                                         const char *data);
STATIC void		dgdm_isdn_flush_channel(BOARD_T *board,
                                                DXBSTRM *channel,
                                                int read_flush,
                                                int write_flush);
STATIC int		dgdm_isdn_drv_ioctl(BOARD_T *board, int device_index,
                                            uchar *data);
STATIC int		dgdm_isdn_openack_cb(DXB *dxb, int channel,
                                             int error_code, DXB_ARG arg);
STATIC int		dgdm_isdn_closeack_cb(DXB *dxb, int channel,
                                              DXB_ARG arg);
STATIC int		dgdm_isdn_pushack_cb(DXB *dxb, int channel,
                                             int error_code, DXB_ARG arg);
STATIC int		dgdm_isdn_popack_cb(DXB *dxb, int channel,
                                            int error_code, DXB_ARG arg);
STATIC int		dgdm_isdn_linkack_cb(DXB *dxb, int channel, int mux_id,
                                             int error_code, DXB_ARG arg);
STATIC int		dgdm_isdn_unlinkack_cb(DXB *dxb, int channel,
                                               int mux_id, int error_code,
                                               DXB_ARG arg);
STATIC int		dgdm_isdn_ioctlack_cb(DXB *dxb, int channel,
                                              int command, int data_len,
                                              int error_code, DXB_ARG arg);
STATIC int		dgdm_isdn_flushack_cb(DXB *dxb, int channel,
                                              int error_code, DXB_ARG arg);
STATIC int		dgdm_isdn_credit_cb(DXB *dxb, int channel, int credits,
                                            DXB_ARG arg);
STATIC int		dgdm_isdn_data_cb(DXB *dxb, int channel,
                                          int data_length, DXB_ARG arg);
STATIC int		dgdm_isdn_cdata_cb(DXB *dxb, int channel, int priority,
                                           int ctrl_length, int data_length,
                                           DXB_ARG arg);
STATIC int		dgdm_isdn_open(struct inode *inode, struct file *file);
STATIC CLOSE_RETURN_T	dgdm_isdn_close(struct inode *inode,struct file *file);
STATIC int		dgdm_isdn_read(READ_PROTO);
STATIC int		dgdm_isdn_write(WRITE_PROTO);
STATIC int		dgdm_isdn_ioctl(struct inode *inode, struct file *file,
                                        unsigned int cmd, unsigned long arg);


/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_read_channel.
 *
 *  Purpose:
 *
 *      Read data from the DxB channel.
 *
 *
 *  Method:
 *
 *      o Copy data from the rx_buffer into the user buffer when available.
 *      o Pend until data is available.
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_read_channel(
	BOARD_T     *board,             /* (i) Board control structure. */
	int         device_index,       /* (i) Device doing the read. */
	char        *data,              /* (i) Data buffer. */
	int         bytes,              /* (i) Byte length of data buffer. */
	loff_t	*ppos			/* (io) ptr to file  position */
)
{
	DXBSTRM *ds;                   /* User's DxB channel. */
	int     control_bytes;         /* Number of control bytes. */
	int     data_bytes;            /* Number of data bytes. */
	int     buffer_bytes;          /* Number of data bytes to copy. */
	int     read_bytes;            /* Bytes written into user buffer. */

	ds = (DXBSTRM *) &board->dxbisdn_chans[device_index];

	if(ds->state != DXBSTRM_OPEN)
		return(-EINVAL);

	/*
	 * Insure the buffer is of a minimum size.
	 */
	if(bytes < sizeof(DXB_STRM_MESSAGE))
		return(-EINVAL);

	/*
	 * User buffer looks like this:
	 *
	 * DXB_STRM_MESSAGE, Control buffer (len != 0), Data Buffer (len != 0).
	 *
	 * Get the Read request.
	 * Only copy in the DXB_STRM_MESSAGE portion of the buffer.
	 */
	COPY_FROM_USER(&ds->rx_msg, data, sizeof(DXB_STRM_MESSAGE));

	/*
	 * Validate the buffer size.
	 */
	if(bytes < (sizeof(DXB_STRM_MESSAGE) + ds->rx_msg.control.maxlen +
		     ds->rx_msg.data.maxlen))
		return(-EINVAL);

	while (1) {
		dxb_lock(board);

		/*
		 * Get any stored DxB requests, and return to the user.
		 */
		if(ds->rx_buffer_bytes > 0) {
			control_bytes = (int) *((ushort *) 
						&ds->rx_buffer[ds->rx_out]);
			data_bytes = (int) *((ushort *) 
					     &ds->rx_buffer[ds->rx_out +
							   sizeof(ushort)]);
			
			read_bytes = control_bytes + data_bytes;

			/*
			 * Update the message structure.
			 */
			ds->rx_msg.control.len = control_bytes;
			ds->rx_msg.data.len = data_bytes;
			
			if((control_bytes > ds->rx_msg.control.maxlen) ||
			    (data_bytes > ds->rx_msg.data.maxlen)) {
				COPY_TO_USER(data, &ds->rx_msg,
					     sizeof(DXB_STRM_MESSAGE));
				dxb_unlock(board);

				return(-ENOSPC);
			}

			RX_HDR_INCR(ds, ds->rx_out);
			ds->rx_buffer_bytes -= ( INTERNAL_COUNTERS );

			/*
			* Copy to user space the message portion of the buffer.
			*/
			COPY_TO_USER(data, &ds->rx_msg,
				     sizeof(DXB_STRM_MESSAGE));
			data = data + sizeof(DXB_STRM_MESSAGE);

			/*
			* Copy the control portion of the message.
			*/
			while (control_bytes) {
				RX_MAX_OUT(ds, control_bytes, buffer_bytes);
				COPY_TO_USER(data, &ds->rx_buffer[ds->rx_out],
					     buffer_bytes);
				data = data + buffer_bytes;
				RX_INCR(ds, ds->rx_out, buffer_bytes);

				control_bytes -= buffer_bytes;
			}

			/*
			 * Adjust data pointer to the end of the control
			 * portion of the buffer.
			 */
			data = data + (ds->rx_msg.control.maxlen - 
				       ds->rx_msg.control.len);

			/*
			* Copy the data portion of the message.
			*/
			while (data_bytes) {
				RX_MAX_OUT(ds, data_bytes, buffer_bytes);
				COPY_TO_USER(data, &ds->rx_buffer[ds->rx_out],
					     buffer_bytes);
				data = data + buffer_bytes;
				RX_INCR(ds, ds->rx_out, buffer_bytes);

				data_bytes -= buffer_bytes;
			}
            
			if(! ds->rx_msg.priority) {
				dxb_return_fixed_cost(board->dxb, ds->dxbchan,
						      1);
				dxb_return_credits(board->dxb, ds->dxbchan,
						   read_bytes);
			}

			dxb_unlock(board);

			/*
			 *	Accumulate Rx stats
			 */
			ds->stat[STAT_StrRxBytes].cnt += read_bytes;
			atomic_add(read_bytes, (atomic_t *) 
				   &ds->stat[STAT_StrTxBytes].accum);

			++ds->stat[STAT_StrRxPackets].cnt;
			atomic_inc((atomic_t *) 
				   &ds->stat[STAT_StrRxPackets].accum);

			*ppos += (read_bytes + sizeof(DXB_STRM_MESSAGE));

			return(read_bytes);
		}

		/*
		* Indicate the user's buffer size awaiting data.
		*/
		ds->user_buffer = data;

		/*
		 * Sleep waiting until the user interrupts (kill signal) or
		 * data arrives from the FEP.
		 */
		ds->read_sleep_count++;

		dxb_unlock(board);
		interruptible_sleep_on(&ds->read_wait);

		ds->user_buffer = NULL;

		if(signal_pending(current)) {
			return(-EINTR);
		}
	} /* while waiting to process data */
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_write_channel.
 *
 *  Purpose:
 *
 *      Write data to the DxB channel.
 *
 *
 *  Method:
 *
 *      o Differentiate between DxB commands.
 *
 *
 *  Returns:
 *
 *      o The appropriate count or error code is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_write_channel(
	BOARD_T     *board,             /* (i) Board control structure. */
	int         device_index,       /* (i) Device doing the write. */
	const char  *data,              /* (i) Data buffer. */
	int         bytes,              /* (i) Byte length of data buffer. */
	loff_t	*ppos			/* (io) ptr to file  position */
)
{
	DXBSTRM *channel;               /* User's DxB channel. */
	DXB_STRM_MESSAGE *msg;          /* User request. */
	int     ret_value;              /* Return value. */

	channel = (DXBSTRM *) &board->dxbisdn_chans[device_index];

	/*
	* Prevent invalid state.
	*/
	if((channel->state != DXBSTRM_OPEN) &&
	    (channel->state != DXBSTRM_OPENING)) {
		DPR(("dgdm_isdn_write_channel: state invalid:%d for "
                     "device:%d\n", (int)channel->state, device_index));
		return(-EINVAL);
	}
    
	/*
	 * First open, need some internal memory for bufferage.
	 */
	if((channel->state == DXBSTRM_OPENING) &&
	    (channel->tx_buffer_size == 0)) {
		/*
		 * Get a minimum of PAGE_SIZE bytes of buffer.
		 */
		if(bytes > PAGE_SIZE)
			channel->tx_buffer_size = bytes;
		else
			channel->tx_buffer_size = PAGE_SIZE;

		channel->tx_buffer = kmalloc(channel->tx_buffer_size,
					     GFP_KERNEL);
		if(channel->tx_buffer == NULL)
			return(-ENOMEM);
	}

	/*
	 * Only can copy in the amount of data specified in the (default)
	 * tx_buffer or the credit amount supplied in the open request.
	 */
	if(bytes > channel->tx_buffer_size) {
		DPR(("dgdm_isdn_write_channel: bytes:%d > tx_buffer_size:%d\n",
		     bytes, channel->tx_buffer_size));
		return(-EINVAL);
	}

	/*
	 * Get the request.
	 */
	COPY_FROM_USER(channel->tx_buffer, data, bytes);

	ret_value = -EINVAL;

	/*
	 * Decode the request.
	 */
	msg = (DXB_STRM_MESSAGE *) channel->tx_buffer;
	switch (msg->command) {
        case DXB_TYPE_OPEN:
		if((msg->control.len > 0) &&
#ifndef RAWDXBDRV
		    (msg->data.len == sizeof(int))
#else
		    ((msg->data.len == sizeof(int)) || channel->very_raw )
#endif
		    )
			ret_value = dgdm_isdn_open_channel(board, channel,
							 msg, data, bytes);
		else
			return(-EINVAL);
		break;
	case DXB_TYPE_CLOSE:
		ret_value = dgdm_isdn_close_channel(board, device_index);
	case DXB_TYPE_DATA:
        case DXB_TYPE_CDATA:
		ret_value = dgdm_isdn_write_data(board, channel, msg);
        case DXB_TYPE_CANCEL:
		/* need command */
		break;
        case DXB_TYPE_PUSH:
		ret_value = dgdm_isdn_push(board, channel, msg, data);
		break;
        case DXB_TYPE_POP:
		ret_value = dgdm_isdn_pop(board, channel, msg, data);
		break;
        case DXB_TYPE_LINK:
		ret_value = dgdm_isdn_link(board, channel, msg, data);
		break;
        case DXB_TYPE_UNLINK:
		ret_value = dgdm_isdn_unlink(board, channel, msg, data);
		break;
        case DXB_TYPE_FLUSH:
		dgdm_isdn_flush_channel(board, channel, 1, 1);
		ret_value = 0 ;
		break;

        default:
		DPR(("dgdm_isdn_write_channel: unknown command:%x, bytes:%d, "
		     "device:%d\n", msg->command,bytes, device_index));
            return(-EINVAL);
	}

	/*
	 * Update the file information.
	 */
	*ppos += bytes;

	return(ret_value);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_open_channel.
 *
 *  Purpose:
 *
 *      Open a DxB channel.
 *
 *
 *  Method:
 *
 *      o Upgrade the tx buffer.
 *      o Allocate the rx buffer.
 *      o Open the DxB channel.
 *      o Set the callback functions.
 *      o Tear down the channel upon error.
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_open_channel(
	BOARD_T     *board,             /* (i) Board control structure. */
	DXBSTRM     *channel,           /* (i) User's DxB channel. */
	DXB_STRM_MESSAGE *msg,          /* (i) User request. */
	const char  *data,              /* (i) Data buffer. */
	int         bytes               /* (i) Byte length of data buffer. */
)
{
	int     open_status;            /* DxB open call status. */
	int     max_credits;            /* Max. allowable credits. */
	uchar   *open_device;           /* STREAMS driver to open on FEP. */
#ifdef RAWDXBDRV
	uchar   *user_data ;		/* DATA for open */
	int     user_data_len ;
#endif

	max_credits = *((int *) (((uchar *) msg) + msg->data.buffer_offset));

	/*
	 * Increment to the next word (4 byte) boundary.
	 */
	max_credits = (max_credits + 3) & ~3;
	DPR(("dgdm_isdn_open_channel channel address 0x%x\n",(uint)channel));

	/*
	 * Upgrade the tx buffer.
	 */
	if(max_credits > channel->tx_buffer_size) {
		kfree(channel->tx_buffer);
		channel->tx_buffer_size = 0;

		channel->tx_buffer = kmalloc((max_credits +
					      sizeof(DXB_STRM_MESSAGE)), 
					     GFP_KERNEL);
		if(channel->tx_buffer == NULL)
			return(-ENOMEM);

		channel->tx_buffer_size = (max_credits +
					   sizeof(DXB_STRM_MESSAGE));

		/*
		 * New tx_buffer, reacquire the request.
		 */
		COPY_FROM_USER(channel->tx_buffer, data, bytes);
		msg = (DXB_STRM_MESSAGE *) channel->tx_buffer;
	}

	/*
	 * Get a user rx buffer (kernel bufferage to cover the max dxb
	 * credits allowed.
	 */
	channel->rx_buffer = kmalloc((max_credits + msg->control.maxlen + 
				      RX_BUFFER_OVERHEAD), GFP_KERNEL);
	if(channel->rx_buffer == NULL) {
		kfree(channel->tx_buffer);
		channel->tx_buffer_size = 0;
		channel->tx_buffer = NULL;

		return(-ENOMEM);
	}

	channel->rx_buffer_size = (max_credits + msg->control.maxlen + 
				   RX_BUFFER_OVERHEAD);
	channel->rx_buffer_bytes = 0;
	channel->rx_in = 0;
	channel->rx_out = 0;

	open_device = (uchar *) (((uchar *) msg) + msg->control.buffer_offset);

#ifdef RAWDXBDRV
	user_data = NULL;
	user_data_len = msg->data.len - sizeof(int) ;
	if(user_data_len)
		user_data = (uchar *) (((uchar *) msg) +
				       msg->data.buffer_offset +
				       sizeof(int)) ;
#endif
	/*
	* Do the open request.
	*/
	for (;;) {
		dxb_lock(board);
#ifndef RAWDXBDRV
		channel->dxbchan = dxb_open(board->dxb, open_device,
					    INTERNAL_COUNTERS, NULL, 0);
#else
        channel->dxbchan = dxb_open(board->dxb, open_device, INTERNAL_COUNTERS,
				    user_data, user_data_len);
#endif

        if(channel->dxbchan >= 0)
		break;

        dxb_unlock(board);

        /*
	 * Memory was not available, wait to try again.
	 *
	 * User can kill the application trying to open the device.
	 */
        if(dgdm_sleep(1))
		return(-EINTR);
	}

	DPR(("DxB channel:%d has credits %d\n",channel->dxbchan,max_credits));

	dxb_return_credits(board->dxb, channel->dxbchan, max_credits);
	dxb_set_credit_return_level(board->dxb, channel->dxbchan,
				    (max_credits / 8));

	dxb_set_openack_cb(board->dxb, channel->dxbchan, dgdm_isdn_openack_cb,
			   (DXB_ARG) channel);
	dxb_set_data_cb(board->dxb, channel->dxbchan, dgdm_isdn_data_cb,
			(DXB_ARG) channel);
	dxb_set_cdata_cb(board->dxb, channel->dxbchan, dgdm_isdn_cdata_cb,
			 (DXB_ARG) channel);
	dxb_set_credit_cb(board->dxb, channel->dxbchan, dgdm_isdn_credit_cb,
			  (DXB_ARG) channel);
	dxb_set_ioctlack_cb(board->dxb, channel->dxbchan, dgdm_isdn_ioctlack_cb,
			    (DXB_ARG) channel);
	dxb_set_pushack_cb(board->dxb, channel->dxbchan, dgdm_isdn_pushack_cb,
			   (DXB_ARG) channel);
	dxb_set_popack_cb(board->dxb, channel->dxbchan, dgdm_isdn_popack_cb,
			  (DXB_ARG) channel);
	dxb_set_linkack_cb(board->dxb, channel->dxbchan, dgdm_isdn_linkack_cb,
			   (DXB_ARG) channel);
	dxb_set_unlinkack_cb(board->dxb, channel->dxbchan,
			     dgdm_isdn_unlinkack_cb, (DXB_ARG) channel);
	dxb_set_flushack_cb(board->dxb, channel->dxbchan, dgdm_isdn_flushack_cb,
			    (DXB_ARG) channel);
	dxb_set_closeack_cb(board->dxb, channel->dxbchan, dgdm_isdn_closeack_cb,
			    (DXB_ARG) channel);

	dxb_unlock(board);

	/*
	 * Wait for the FEP to acknowledge the open request.
	 */
	interruptible_sleep_on(&channel->openack_wait);

	if(signal_pending(current))
		open_status = -EINTR;
	else
		if(channel->error_code)
			open_status = -ENODEV;
		else
			open_status = channel->dxbchan;

	if(open_status < 0) {
		kfree(channel->tx_buffer);
		kfree(channel->rx_buffer);
		channel->tx_buffer = NULL;
		channel->rx_buffer = NULL;

		channel->tx_buffer_size = 0;
		channel->rx_buffer_size = 0;

		channel->state = DXBSTRM_IDLE;
	} else {
		channel->state = DXBSTRM_OPEN;
	}

	return(open_status);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_close_channel.
 *
 *  Purpose:
 *
 *      Close a DxB channel.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_close_channel(
	BOARD_T     *board,             /* (i) Board control structure. */
	int         device_index        /* (i) Device doing the close. */
)
{
	DXBSTRM *channel;              /* User's DxB channel. */
	int     close_status;          /* Return value. */

	channel = (DXBSTRM *) &board->dxbisdn_chans[device_index];

	if((channel->state != DXBSTRM_OPEN) &&
	    (channel->state != DXBSTRM_OPENING))
		return(-EINVAL);
    
	channel->state = DXBSTRM_CLOSE;

#ifndef RAWDXBDRV
	dgdm_isdn_flush_channel(board, channel, 1, 1);
#else
	if(!channel->very_raw)
		dgdm_isdn_flush_channel(board, channel, 1, 1);
#endif

	for (;;) {
		/*
		 * Request a DxB close.
		 */
		dxb_lock(board);
		close_status = dxb_close(board->dxb, channel->dxbchan);
		dxb_unlock(board);

		if(close_status == 0) {
			interruptible_sleep_on(&channel->closeack_wait);
			break;
		} else {
			/*
			 * Memory was not available, wait to try again.
			 *
			 * User can kill the application trying to
			 * close the device.
			 */
			if(dgdm_sleep(1))
				break;
		}
	}

	dxb_lock(board);
	kfree(channel->tx_buffer);
	kfree(channel->rx_buffer);

	channel->tx_buffer = NULL;
	channel->rx_buffer = NULL;
	channel->tx_buffer_size = 0;
	channel->rx_buffer_size = 0;

	channel->state = DXBSTRM_IDLE;
	dxb_unlock(board);

	return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_write_data.
 *
 *  Purpose:
 *
 *      Send write data to the FEP.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Returns the number of bytes written or the cause for the error.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_write_data(
	BOARD_T     *board,             /* (i) Board control structure. */
	DXBSTRM     *channel,           /* (i) User's DxB channel. */
	DXB_STRM_MESSAGE *msg           /* (i) User request. */
)

{
	int     dxb_space;              /* Space available in DxB? */

	/*
	 * Wait until the DxB interface has room for the request.
	 */
	for (;;) {
		dxb_lock(board);
		dxb_space = dxb_space_avail(board->dxb, channel->dxbchan,
					    msg->control.len, msg->data.len, 
					    msg->priority);

		if(dxb_space == DXB_OK) {
			break;
		} else {
			if(dxb_space == DXB_ENOCRED) {
				channel->write_credits_needed = 
					msg->control.len + msg->data.len;
				channel->write_sleep_count++;
				/*
				 * Wait for more room in the DxB buffer.
				 */
				dxb_unlock(board);
				interruptible_sleep_on(&channel->write_wait);
				channel->write_credits_needed = 0;
				if(signal_pending(current)) {
					return(-EINTR);
				}
			} else {
				if(dxb_space == DXB_ENOSPACE) {
					dxb_unlock(board);

					/*
					 * Memory was not available, wait
					 * to try again.
					 *
					 * User can kill the application
					 * trying to open the device.
					 */
					if( dgdm_sleep(1) ) {
						return(-EINTR);
					}
				}
				else {
					dxb_unlock(board);
					DPR(("dgdm_isdn_write_data bad status "
					     "from dxb_space_avail():%d\n",
					     dxb_space));
					return(-EINVAL);
				}
			}
		}
	}

	if(msg->control.len) {
		(void) dxb_cdata(board->dxb, channel->dxbchan,
				 msg->priority, 
				 msg->control.len, msg->data.len);
		(void) dxb_write_ctrl_block(board->dxb, 
					(uchar *) (((uchar *) msg) + 
						   msg->control.buffer_offset),
					    msg->control.len);
	} else {
		(void) dxb_data(board->dxb, channel->dxbchan,
				msg->data.len);
	}

	if(msg->data.len) {
		(void) dxb_write_data_block(board->dxb, 
					(uchar *) (((uchar *) msg) +
						   msg->data.buffer_offset),
					    msg->data.len);
	}

	dxb_write_done(board->dxb);
	dxb_unlock(board);

	/*
	 *	Accumulate Tx stats
	 */
	channel->stat[STAT_StrTxBytes].cnt += msg->data.len;
	atomic_add(msg->data.len,
		   (atomic_t *) &channel->stat[STAT_StrTxBytes].accum);

	++channel->stat[STAT_StrTxPackets].cnt;
	atomic_inc((atomic_t *) &channel->stat[STAT_StrTxPackets].accum);

	return(msg->data.len + msg->control.len);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_push.
 *
 *  Purpose:
 *
 *      Send module name to push to the FEP.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Returns the number of bytes written or the cause for the error.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_push(
	BOARD_T     *board,             /* (i) Board control structure. */
	DXBSTRM     *channel,           /* (i) User's DxB channel. */
	DXB_STRM_MESSAGE *msg,          /* (i) User request. */
	const char  *data               /* (i) Data buffer. */
)
{
	int     push_status;
	uchar   *module_id;
	uchar   *module_info;
	uchar   *error_code;

	/*
	 * Get the module name and information.
	 */
	module_id = (uchar *) (((uchar *) msg) + msg->control.buffer_offset);
	module_info = (uchar *) (((uchar *) msg) + msg->data.buffer_offset);
	error_code = (uchar *) (((uchar *) msg) + msg->data.buffer_offset);

	for (;;) {
		/*
		* Request a DxB push.
		*/
		dxb_lock(board);
		push_status = dxb_push(board->dxb, channel->dxbchan, module_id,
				       module_info, msg->data.len);
		dxb_unlock(board);

		if(push_status == 0) {
			interruptible_sleep_on(&channel->pushack_wait);
			break;
		} else {
			/*
			 * Memory was not available, wait to try again.
			 *
			 * User can kill the application trying to close
			 * the device.
			 */
			if(dgdm_sleep(1))
				break;
		}
	}

	*((int *) error_code) = channel->error_code;
	msg->data.len = sizeof(int);

	/*
	 * THIS IS A KLUDGE.  IT WILL DO FOR NOW!!!
	 */
	COPY_TO_USER((char *) data, (char *) msg, 
		     (msg->data.len+ msg->control.maxlen +
		      sizeof(DXB_STRM_MESSAGE)));

	return(msg->data.len);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_pop.
 *
 *  Purpose:
 *
 *      Send pop message to the FEP.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Returns the number of bytes written or the cause for the error.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int
dgdm_isdn_pop(
	BOARD_T     *board,             /* (i) Board control structure. */
	DXBSTRM     *channel,           /* (i) User's DxB channel. */
	DXB_STRM_MESSAGE *msg,          /* (i) User request. */
	const char  *data               /* (i) Data buffer. */
)
{
	int     pop_status;
	uchar   *error_code;

	/*
	 * Get the module name and information.
	 */
	error_code = (uchar *) (((uchar *) msg) + msg->data.buffer_offset);

	for (;;) {
		/*
		 * Request a DxB push.
		 */
		dxb_lock(board);
		pop_status = dxb_pop(board->dxb, channel->dxbchan);
		dxb_unlock(board);

		if(pop_status == 0) {
			interruptible_sleep_on(&channel->popack_wait);
			break;
		} else {
			/*
			 * Memory was not available, wait to try again.
			 *
			 * User can kill the application trying to close
			 * the device.
			 */
			if(dgdm_sleep(1))
				break;
		}
	}

	*((int *) error_code) = channel->error_code;
	msg->data.len = sizeof(int);

	/*
	 * THIS IS A KLUDGE.  IT WILL DO FOR NOW!!!
	 */
	COPY_TO_USER((char *) data, (char *) msg, 
		     (msg->data.len + msg->control.maxlen +
		      sizeof(DXB_STRM_MESSAGE)));

	return(msg->data.len);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_link.
 *
 *  Purpose:
 *
 *      Send link message to the FEP.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Returns the number of bytes written or the cause for the error.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_link(
        BOARD_T     *board,             /* (i) Board control structure. */
        DXBSTRM     *channel,           /* (i) User's DxB channel. */
        DXB_STRM_MESSAGE *msg,          /* (i) User request. */
        const char  *data               /* (i) Data buffer. */
)
{
        int     link_status;
        int     *other_channel;
        int     *mux_id;
        int     *error_code;

        /*
         * Setup the addresses of the parameters.
         */
        other_channel = (int *) (((uchar *) msg) + msg->data.buffer_offset);
        error_code = (int *) (((uchar *) msg) + msg->data.buffer_offset);
        mux_id = (int *) (error_code + 1);

        DPR(("dgdm_isdn_link channel:%d other_channel:%d\n",
             channel->dxbchan, *other_channel));

        for (;;) {
                /*
                * Request a DxB link.
                */
                dxb_lock(board);
                link_status = dxb_link(board->dxb, channel->dxbchan,
                                       *other_channel);
                dxb_unlock(board);

                if(link_status == 0) {
                        interruptible_sleep_on(&channel->linkack_wait);
                        break;
                } else {
                        /*
                        * Memory was not available, wait to try again.
                        *
                        * User can kill the application trying to close
                        * the device.
                        */
                        if(dgdm_sleep(1))
                                break;
                }
        }

        *error_code = channel->error_code;
        *mux_id = channel->mux_id;
        msg->data.len = sizeof(int)*2;

        /*
         * THIS IS A KLUDGE.  IT WILL DO FOR NOW!!!
         */
        COPY_TO_USER((char *) data, (char *) msg, 
                     (msg->data.len + msg->control.maxlen + 
                      sizeof(DXB_STRM_MESSAGE)));

        return(msg->data.len);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_unlink.
 *
 *  Purpose:
 *
 *      Send unlink message to the FEP.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Returns the number of bytes written or the cause for the error.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_unlink(
        BOARD_T     *board,             /* (i) Board control structure. */
        DXBSTRM     *channel,           /* (i) User's DxB channel. */
        DXB_STRM_MESSAGE *msg,          /* (i) User request. */
        const char  *data               /* (i) Data buffer. */
)
{
        int     unlink_status;
        int     *mux_id;
        int     *mux_id_return;
        int     *error_code;

        /*
         * Setup the addresses of the parameters.
         */
        mux_id = (int *) (((uchar *) msg) + msg->data.buffer_offset);
        error_code = (int *) (((uchar *) msg) + msg->data.buffer_offset);
        mux_id_return = (int *) (error_code + 1);

        for (;;) {
                /*
                 * Request a DxB link.
                 */
                dxb_lock(board);
                unlink_status = dxb_unlink(board->dxb, channel->dxbchan,
                                           *mux_id);
                dxb_unlock(board);

                if(unlink_status == 0) {
                        interruptible_sleep_on(&channel->unlinkack_wait);
                        break;
                } else {
                        /*
                         * Memory was not available, wait to try again.
                         *
                         * User can kill the application trying to close
                         * the device.
                         */
                        if(dgdm_sleep(1))
                                break;
                }
        }

        *error_code = channel->error_code;
        *mux_id_return = channel->mux_id;
        msg->data.len = sizeof(int)*2;

        /*
         * THIS IS A KLUDGE.  IT WILL DO FOR NOW!!!
         */
        COPY_TO_USER((char *) data, (char *) msg, 
                     (msg->data.len + msg->control.maxlen +
                      sizeof(DXB_STRM_MESSAGE)));

        return(msg->data.len);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_flush_channel.
 *
 *  Purpose:
 *
 *      Send flush to the FEP.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Ignores any and all errors!!!
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/10/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC void dgdm_isdn_flush_channel(
        BOARD_T     *board,             /* (i) Board control structure. */
        DXBSTRM     *channel,           /* User's DxB channel. */
        int         read_flush,         /* (i) Flush Read. */
        int         write_flush         /* (i) Flush Write. */
)
{
        int     flush_status;

        for (;;) {
                /*
                 * Request a DxB flush.
                 */
                dxb_lock(board);
                flush_status = dxb_flush(board->dxb, channel->dxbchan,
                                         read_flush, write_flush);
                dxb_unlock(board);

                if(flush_status == 0) {
#if(0)
                        sleep_on(&channel->flushack_wait);
#else
                        interruptible_sleep_on(&channel->flushack_wait);
#endif
                        break;
                } else {
                        /*
                         * Memory was not available, wait to try again.
                         *
                         * User can kill the application trying to close
                         * the device.
                         */
                        if(dgdm_sleep(1))
                                break;
                }
        }
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_drv_ioctl.
 *
 *  Purpose:
 *
 *      Send IOCTL to the FEP.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Returns the number of bytes written or the cause for the error.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/30/98  1.0.0.0  J. Ruf         IOCTL request from driver directly.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_drv_ioctl(
        BOARD_T     *board,             /* (i) Board control structure. */
        int         device_index,       /* (i) Device doing the write. */
        uchar       *data               /* (i) Data buffer. */
)
{
        DXBSTRM *channel;               /* User's DxB channel. */
        int     dxb_space;              /* Space available in DxB? */
        strioctl str_ioctl;

        channel = (DXBSTRM *) &board->dxbisdn_chans[device_index];

        /*
         * Prevent invalid state.
         */
        if(channel->state != DXBSTRM_OPEN) {
                DPR(("dgdm_isdn_drv_write: state invalid:%d for device:%d\n",
                     (int)channel->state, device_index));
                return(-EINVAL);
        }

        /*
         * Get the request.
         */
        COPY_FROM_USER(&str_ioctl, data, sizeof(strioctl));

        if((str_ioctl.ic_dp) && (str_ioctl.ic_len)) {
                COPY_FROM_USER(channel->tx_buffer, str_ioctl.ic_dp,
                               str_ioctl.ic_len);
        }

        /*
         * Now do the DxB operation.
         */

        /*
         * Wait until the DxB interface has room for the request.
         */
        for (;;) {
                dxb_lock(board);
                dxb_space = dxb_ioctl_space_avail(board->dxb, channel->dxbchan,
                                                  str_ioctl.ic_len);

                if(dxb_space == DXB_OK) {
                        break;
                } else if(dxb_space == DXB_ENOCRED) {
                        DPR(("dxbisdb_drv_ioctl:  WAITING ON CREDITS!\n"));
                        channel->ioctl_credits_needed = str_ioctl.ic_len;

                        /*
                         * Wait for more room in the DxB buffer.
                         */
                        dxb_unlock(board);
                        interruptible_sleep_on(&channel->ioctl_wait);
                        if(signal_pending(current)) {
                                return(-EINTR);
                        }
                } else if(dxb_space == DXB_ENOSPACE) {
                        dxb_unlock(board);

                        /*
                         * Memory was not available, wait to try again.
                         *
                         * User can kill the application trying to open
                         * the device.
                         */
                        if(dgdm_sleep(1)) {
                                return(-EINTR);
                        }
                } else {
                        dxb_unlock(board);
                        DPR(("dgdm_isdn_drv_ioctl bad status from "
                             "dxb_ioctl_space_avail():%d\n", dxb_space));
                        return(-EINVAL);
                }
        }

        (void) dxb_ioctl(board->dxb, channel->dxbchan, str_ioctl.ic_cmd,
                         str_ioctl.ic_len);
        (void) dxb_write_ioctl_block(board->dxb, (uchar *) channel->tx_buffer,
                                     str_ioctl.ic_len);
        dxb_write_done(board->dxb);
        dxb_unlock(board);

        /*
         * Wait for the IOCTL to finish.
         */
        for (;;) {
                /*
                 * Sleep waiting until the user interrupts (kill signal) or
                 * data arrives from the FEP.
                 */
                interruptible_sleep_on(&channel->ioctl_wait);
                if(signal_pending(current))
                        return(-EINTR);
                else
                        break;
        }

        dxb_lock(board);

        /*
         * NOTE:
         *
         * The callback function places all the return information about the
         * IOCTL in the channel structure.
         */

        /*
         * Get the return data.
         */

        /*
         * Return the results.
         */
        if((str_ioctl.ic_dp) && (channel->ioctl_data_len)) {
                COPY_TO_USER(str_ioctl.ic_dp, channel->tx_buffer,
                             channel->ioctl_data_len);
        }
        str_ioctl.ic_len = channel->ioctl_data_len;
        COPY_TO_USER(data, &str_ioctl, sizeof(strioctl));

        dxb_unlock(board);

        return(channel->ioctl_errno);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_openack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB open request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_openack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         error_code,         /* (i) Callback return value. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *ds;                /* Driver channel structure. */

        ds = (DXBSTRM *) arg;

        ds->error_code = error_code;

        wake_up_interruptible(&ds->openack_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_closeack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB close request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_closeack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)

{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

        wake_up_interruptible(&drv_channel->closeack_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_pushack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB push request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_pushack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         error_code,         /* (i) Return error code. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

        drv_channel->error_code = error_code;
        wake_up_interruptible(&drv_channel->pushack_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_popack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB pop request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_popack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         error_code,         /* (i) Return error code. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

        drv_channel->error_code = error_code;
        wake_up_interruptible(&drv_channel->popack_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_linkack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB link request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_linkack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         mux_id,             /* (i) MUX ID. */
        int         error_code,         /* (i) Return error code. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

        drv_channel->error_code = error_code;
        drv_channel->mux_id = mux_id;
        wake_up_interruptible(&drv_channel->linkack_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_unlinkack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB unlink request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    01/15/98  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_unlinkack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         mux_id,             /* (i) MUX ID. */
        int         error_code,         /* (i) Return error code. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

        drv_channel->error_code = error_code;
        drv_channel->mux_id = mux_id;
        wake_up_interruptible(&drv_channel->unlinkack_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_ioctlack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB ioctl request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_ioctlack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         command,            /* (i) DxB IOCTL command. */
        int         data_len,           /* (i) Amount of data to copy. */
        int         error_code,         /* (i) Return error code. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

        drv_channel->ioctl_errno = error_code;
        drv_channel->ioctl_data_len = data_len;

        if(data_len) {
                dxb_read_data_block(dxb, (uchar *) drv_channel->tx_buffer,
                                    data_len, data_len);

                /*
                 * Finish IOCTL request.
                 */
                dxb_read_done(dxb);
        }

        wake_up_interruptible(&drv_channel->ioctl_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_flushack_cb.
 *
 *  Purpose:
 *
 *      Finish the DxB flush request.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_flushack_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         error_code,         /* (i) Return error code. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

        if(error_code != DXB_OK)
                DPR(("dgdm_isdn_flushack_cb error:%d\n",error_code));

        wake_up_interruptible(&drv_channel->flushack_wait);

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_credit_cb.
 *
 *  Purpose:
 *
 *      Restart pended/sleeping user write.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_credit_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         credits,            /* (i) Available credits. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *drv_channel;       /* Driver channel structure. */

        drv_channel = (DXBSTRM *) arg;

#if 0
        DPR(("dxisdn_credit_cb channel:%d credits:%d\n", channel, credits));
        DPR(("dxisdn_credit_cb write credits needed:%d ioctl credits "
             "needed:%d\n", drv_channel->write_credits_needed,
             drv_channel->ioctl_credits_needed));
#endif

        if(drv_channel->write_credits_needed) {
                drv_channel->write_wake_count++;
                wake_up_interruptible(&drv_channel->write_wait);
        }

        if(drv_channel->ioctl_credits_needed) {
                drv_channel->ioctl_credits_needed = 0;
                wake_up_interruptible(&drv_channel->ioctl_wait);
        }

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_data_cb.
 *
 *  Purpose:
 *
 *      Process received read data.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_data_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         data_length,        /* (i) Length of data. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM     *ds;               	/* Driver channel structure. */
        int         buffer_bytes;	/* Bytes to copy into buffer. */

        ds = (DXBSTRM *) arg;

        /*
         *	Accumulate Rx stats
         */
        ds->stat[STAT_StrRxBytes].cnt += data_length;
        atomic_add(data_length, (atomic_t *) &ds->stat[STAT_StrTxBytes].accum);

        ++ds->stat[STAT_StrRxPackets].cnt;
        atomic_inc((atomic_t *) &ds->stat[STAT_StrRxPackets].accum);

        /*
         * Put the data into the temporary rx_buffer.
         * Insure that the data will not overflow the temporary buffer.
         */
        RX_BUFFER_FREE(ds, buffer_bytes);
        if(buffer_bytes < (data_length+INTERNAL_COUNTERS)) {
                ds->rx_dropped++;

                dxb_read_done(dxb);

                dxb_return_fixed_cost(dxb, channel, 1);
                dxb_return_credits(dxb, channel, data_length);

                DPR(("dgdm_isdn_data_cb channel:%d data_length:%d DROPPED\n",
                     channel,data_length));
                DPR(("dgdm_isdn_data_cb rx_in:%d rx_out:%d rx size:%d\n",
                     ds->rx_in,ds->rx_out,ds->rx_buffer_size));

                /*
                 * CAUTION:
                 *
                 * User waiting on data will never enter this path UNLESS
                 * MASSIVE credit FAILURE in DXB interface.
                 *
                 * If failure, then possible hanging user application.
                 */
                return(0);
        }

        *((ushort *) &ds->rx_buffer[ds->rx_in]) = (ushort) 0;
        *((ushort *) &ds->rx_buffer[ds->rx_in + sizeof(ushort)]) = 
                (ushort) data_length;
        RX_HDR_INCR(ds, ds->rx_in);
        ds->rx_buffer_bytes += (  INTERNAL_COUNTERS  );
    
        while (data_length) {
                RX_MAX_IN(ds, ds->rx_in, data_length, buffer_bytes);
                dxb_read_data_block(dxb, &ds->rx_buffer[ds->rx_in], 
                                    buffer_bytes, buffer_bytes);
                RX_INCR(ds, ds->rx_in, buffer_bytes);
        
                data_length -= buffer_bytes;
        }

        dxb_read_done(dxb);

        /*
         * Wake the user.
         */
        if(ds->user_buffer != NULL) {
                ds->read_wake_count++;
                wake_up_interruptible(&ds->read_wait);
        }

        return(0);
}



/*
 *----------------------------------------------------------------------
 *  Function:  dgdm_isdn_cdata_cb.
 *
 *  Purpose:
 *
 *      Process received read control and data.
 *
 *
 *  Method:
 *
 *      o .
 *
 *
 *  Returns:
 *
 *      o Nothing is returned.
 *
 *
 *  Revision History:
 *
 *      Date   Revision  Author         Modification Notes
 *
 *    12/29/97  1.0.0.0  J. Ruf         Initial Release.
 *
 *----------------------------------------------------------------------
 */
STATIC int dgdm_isdn_cdata_cb(
        DXB         *dxb,               /* (i) DxB channel information. */
        int         channel,            /* (i) DxB channel number. */
        int         priority,           /* (i) DxB message priority. */
        int         ctrl_length,        /* (i) Length of control contents. */
        int         data_length,        /* (i) Length of data. */
        DXB_ARG     arg                 /* (i) Driver DxB channel structure. */
)
{
        DXBSTRM *ds;                    /* Driver channel structure. */
        int     buffer_bytes;           /* Bytes to copy into buffer. */

        ds = (DXBSTRM *) arg;

        /*
         * Put the data into the temporary rx_buffer.
         * Insure that the data will not overflow the temporary buffer.
         */
        RX_BUFFER_FREE(ds, buffer_bytes);
        if(buffer_bytes < (data_length+ctrl_length+( INTERNAL_COUNTERS ))) {
                ds->rx_dropped++;

                dxb_read_done(dxb);

                if(! priority) {
                        dxb_return_credits(dxb, channel, 
                                           data_length + ctrl_length);
                        dxb_return_fixed_cost(dxb, channel, 1);
                }

                DPR(("dgdm_isdn_cdata_cb channel:%d ctrl_length:%d "
                     "data_length:%d DROPPED\n", channel, ctrl_length,
                     data_length));
                DPR(("dgdm_isdn_cdata_cb rx_in:%d rx_out:%d rx size:%d\n",
                     ds->rx_in, ds->rx_out, ds->rx_buffer_size));

                /*
                 * CAUTION:
                 *
                 * User waiting on data will never enter this path UNLESS
                 * MASSIVE credit FAILURE in DXB interface.
                 *
                 * If failure, then possible hanging user application.
                 */
                return(0);
        }

        /*
         * Minimize the amount of buffer wrap going on.
         */
        if(ds->rx_in == ds->rx_out) {
                ds->rx_in = ds->rx_out = 0;
        }

        *((ushort *) &ds->rx_buffer[ds->rx_in]) = (ushort) ctrl_length;
        *((ushort *) &ds->rx_buffer[ds->rx_in +sizeof(ushort)]) = 
                (ushort) data_length;
        RX_HDR_INCR(ds, ds->rx_in);
        ds->rx_buffer_bytes += ( INTERNAL_COUNTERS );
    
        while (ctrl_length) {
                RX_MAX_IN(ds, ds->rx_in, ctrl_length, buffer_bytes);
                dxb_read_ctrl_block(dxb, &ds->rx_buffer[ds->rx_in], 
                                    buffer_bytes, buffer_bytes);
                RX_INCR(ds, ds->rx_in, buffer_bytes);
        
                ctrl_length -= buffer_bytes;
        }

        while (data_length) {
                RX_MAX_IN(ds, ds->rx_in, data_length, buffer_bytes);
                dxb_read_data_block(dxb, &ds->rx_buffer[ds->rx_in], 
                                    buffer_bytes, buffer_bytes);
                RX_INCR(ds, ds->rx_in, buffer_bytes);
        
                data_length -= buffer_bytes;
        }

        dxb_read_done(dxb);

        /*
         * Wake the user.
         */
        if(ds->user_buffer) {
                ds->read_wake_count++;
                wake_up_interruptible(&ds->read_wait);
        }

        return(0);
}




/************************************************************************
*
*	open/close/read/write/ioctl for raw ISDN device
*
*************************************************************************/

#define	minor2isdn(MINOR)	( (MINOR) & 0x3f )

/*
 * Open the raw ISDN device
 */
STATIC int dgdm_isdn_open(struct inode *inode, struct file *file)
{
	unsigned int	minor = MINOR(inode->i_rdev);
	int		board = minor2board(minor);
	int		chan = minor2isdn(minor);
	BOARD_T	*brd =  dgdm_Board[board];

        if(!brd)
                return(-ENODEV);

	if(board >= dgdm_NumBoards)
		return(-ENXIO);

	if(brd->dxbisdn_chans == NULL)
		return(-ENXIO);	/* Illegal */

	if(chan >= NISDN)
		return(-ENXIO);

	if(brd->dxbisdn_chans[chan].state != DXBSTRM_IDLE) {
		/*
		 * The requested device is already in use.
		 */
		return(-EBUSY);
	} else {
		/*
		 * Reserve a device to open a DxB channel
		 * on a subsequent write request.
		 */
		brd->dxbisdn_chans[chan].state = DXBSTRM_OPENING;
#ifdef RAWDXBDRV
		brd->dxbisdn_chans[chan].very_raw = 0 ;
#endif
	}

	MOD_INC_USE_COUNT;

	return(0);
}




/*
 * Close the raw ISDN device
 */
STATIC CLOSE_RETURN_T dgdm_isdn_close(struct inode *inode, struct file *file)
{
	unsigned int	minor = MINOR(inode->i_rdev);
	int		board = minor2board(minor);
	int		chan = minor2isdn(minor);
	BOARD_T	*brd =  dgdm_Board[board];

        if(!brd)
                return(-ENODEV);

	if(dgdm_tty && dgdm_isdn) {
		dgdm_isdn_close_channel(brd, chan);
	}

        MOD_DEC_USE_COUNT;

	CLOSE_RETURN(0);
}




/*
 * Read from the raw ISDN device
 */
STATIC int dgdm_isdn_read(READ_PROTO)
{
	unsigned int	minor = MINOR(PARM_I_RDEV);
	int		board = minor2board(minor);
	int		chan = minor2isdn(minor);
	BOARD_T		*brd = dgdm_Board[board];
#if LINUX_VERSION_CODE < 0x20100
        loff_t		*ppos = &file->f_pos;
#endif

        if(!brd)
                return(-ENODEV);

	if(count < 0)
		return(-EINVAL);

	return(dgdm_isdn_read_channel(brd, chan, buf, count, ppos));
}




/*
 * Write to the ram/control/debug device
 */
STATIC int dgdm_isdn_write(WRITE_PROTO)
{
	unsigned int	minor = MINOR(PARM_I_RDEV);
	int		board = minor2board(minor);
	int		chan = minor2isdn(minor);
	BOARD_T		*brd = dgdm_Board[board];
#if LINUX_VERSION_CODE < 0x20100
        loff_t		*ppos = &file->f_pos;
#endif

        if(!brd)
                return(-ENODEV);

	if(count < 0)
		return(-EINVAL);
	if(count == 0)
		return(0);

	return(dgdm_isdn_write_channel(brd, chan, buf, count, ppos));
}




/*
 * ioctl() the raw ISDN device.
 */
STATIC int dgdm_isdn_ioctl(struct inode *inode, struct file *file,
                           unsigned int cmd, unsigned long arg)
{
	unsigned int	minor = MINOR(inode->i_rdev);
	int		board = minor2board(minor);
	int		chan = minor2isdn(minor);
	BOARD_T		*brd = dgdm_Board[board];

        if(!brd)
                return(-ENODEV);

#ifdef RAWDXBDRV
	if(cmd == I_VERYRAW) {
		brd->dxbisdn_chans[chan].very_raw = 1 ;
		return(0) ;
	}
#endif
	if(cmd != I_STR)
		return(-EINVAL);

	return(dgdm_isdn_drv_ioctl(brd, chan, (uchar *) arg));
}



#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
/*
 * File operations permitted on major+3
 */
GLOBAL struct file_operations dgdm_BoardIsdnFops = {
        owner:		THIS_MODULE,
        read:		dgdm_isdn_read,
	write:	        dgdm_isdn_write,
	ioctl:	        dgdm_isdn_ioctl,
	open:	        dgdm_isdn_open,
        release:	dgdm_isdn_close,
};
#else
/*
 * File operations permitted on major+3
 */
GLOBAL struct file_operations dgdm_BoardIsdnFops = {
	NULL,					/* lseek */
	dgdm_isdn_read,				/* read */
	dgdm_isdn_write,			/* write */
	NULL,					/* readdir */
	NULL,					/* select/poll */
	dgdm_isdn_ioctl,			/* ioctl */
	NULL,					/* mmap */
	dgdm_isdn_open,				/* open */
#if LINUX_VERSION_CODE >= 0x20100
	NULL,					/* flush */
#endif
	dgdm_isdn_close,			/* close/release */
	NULL,					/* fsync */
#if LINUX_VERSION_CODE >= 0x20100
	NULL,					/* fasync */
	NULL,					/* check_media_change */
	NULL,					/* revalidate */
	NULL,					/* lock */
#endif
};
#endif

