/* ex:se ts=8 sw=4 sm: */ /* -*- linux-c -*- */

/*
 * Copyright 1999 by Digi International (www.digi.com)
 *    Jeff Randall <Jeff_Randall@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the 
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define __NO_VERSION__
#include "dgdm_driver.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
# include <linux/vmalloc.h>
#endif

/* file level globals */
static char		*dgdm_trcbuf;	    /* the ringbuffer */

# if defined(TRC_TO_KMEM)
static int		 dgdm_trcbufi = 0;  /* index of the tilde at the end of */
#endif

extern int		 dgdm_trcbuf_size;  /* size of the ringbuffer */

# if defined(TRC_TO_KMEM) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
static spinlock_t	 dgdm_tracef_lock = SPIN_LOCK_UNLOCKED;
#endif


#if !defined(TRC_TO_KMEM) && !defined(TRC_TO_CONSOLE)
GLOBAL void dgdm_tracef(const char *fmt, ...) {
	return;
}

#else /* !defined(TRC_TO_KMEM) && !defined(TRC_TO_CONSOLE) */

GLOBAL void dgdm_tracef(const char *fmt, ...) {
	va_list	         ap;
	char  	         buf[TRC_MAXMSG+1];
	size_t		 lenbuf;
	int		 i;
        static int	 failed = FALSE;
# if defined(TRC_TO_KMEM)
        unsigned long	 flags;
#endif

        if(failed)
                return;
# if defined(TRC_TO_KMEM)
        DGDM_LOCK(dgdm_tracef_lock, flags);
#endif

	/* Format buf using fmt and arguments contained in ap. */
	va_start(ap, fmt);
	i = vsprintf(buf, fmt,  ap);
	va_end(ap);
	lenbuf = strlen(buf);

# if defined(TRC_TO_KMEM)
	{
		static int	 initd=0;

		/*
		 * Now, in addition to (or instead of) printing this stuff out
		 * (which is a buffered operation), also tuck it away into a
		 * corner of memory which can be examined post-crash in kdb.
		 */
		if (!initd) {
                        dgdm_trcbuf = (char *) vmalloc(dgdm_trcbuf_size);
                        if(!dgdm_trcbuf) {
                                failed = TRUE;
                                printk("dgdm: tracing init failed!\n");
                                return;
                        }

                        memset(dgdm_trcbuf, '\0',  dgdm_trcbuf_size);
                        dgdm_trcbufi = 0;
			initd++;

			printk("dgdm: tracing enabled - " TRC_DTRC 
                               " 0x%lx 0x%x\n",
			       (unsigned long)dgdm_trcbuf, 
			       dgdm_trcbuf_size);
		}

#  if defined(TRC_ON_OVERFLOW_WRAP_AROUND)
		/*
		 * This is the less CPU-intensive way to do things.  We simply
		 * wrap around before we fall off the end of the buffer.  A 
		 * tilde (~) demarcates the current end of the trace.
		 *
		 * This method should be used if you are concerned about race
		 * conditions as it is less likely to affect the timing of
		 * things.
		 */

		if (dgdm_trcbufi + lenbuf >= dgdm_trcbuf_size) {
			/* We are wrapping, so wipe out the last tilde. */
			dgdm_trcbuf[dgdm_trcbufi] = '\0';
			/* put the new string at the beginning of the buffer */
			dgdm_trcbufi = 0;
		}

		strcpy(&dgdm_trcbuf[dgdm_trcbufi], buf);	
		dgdm_trcbufi += lenbuf;
		dgdm_trcbuf[dgdm_trcbufi] = '~';

#  elif defined(TRC_ON_OVERFLOW_SHIFT_BUFFER)
		/*
		 * This is the more CPU-intensive way to do things.  If we
		 * venture into the last 1/8 of the buffer, we shift the 
		 * last 7/8 of the buffer forward, wiping out the first 1/8.
		 * Advantage: No wrap-around, only truncation from the
		 * beginning.
		 *
		 * This method should not be used if you are concerned about
		 * timing changes affecting the behaviour of the driver (ie,
		 * race conditions).
		 */
		strcpy(&dgdm_trcbuf[dgdm_trcbufi], buf);
		dgdm_trcbufi += lenbuf;
		dgdm_trcbuf[dgdm_trcbufi] = '~';
		dgdm_trcbuf[dgdm_trcbufi+1] = '\0';

		/* If we're near the end of the trace buffer... */
		if (dgdm_trcbufi > (dgdm_trcbuf_size/8)*7) {
			/* Wipe out the first eighth to make some more room. */
			strcpy(dgdm_trcbuf, &dgdm_trcbuf[dgdm_trcbuf_size/8]);
			dgdm_trcbufi = strlen(dgdm_trcbuf)-1;
			/* Plop overflow message at the top of the buffer. */
			bcopy(TRC_OVERFLOW, dgdm_trcbuf, strlen(TRC_OVERFLOW));
		}
#  else
#   error "TRC_ON_OVERFLOW_WRAP_AROUND or TRC_ON_OVERFLOW_SHIFT_BUFFER?"
#  endif
	}
        DGDM_UNLOCK(dgdm_tracef_lock, flags);
# endif /* defined(TRC_TO_KMEM) */
}

#endif /* !defined(TRC_TO_KMEM) && !defined(TRC_TO_CONSOLE) */




/*
 * dgdm_tracer_free()
 *
 *
 */
GLOBAL void dgdm_tracer_free(void) {

        if(dgdm_trcbuf)
                vfree(dgdm_trcbuf);
}
