/* -*- linux-c -*- (autoset linux indent style in emacs) */

/*
 * Copyright 1999 by Digi International (www.digi.com)
 *    Jeff Randall <Jeff_Randall@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the 
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*****************************************************************************
 *
 *  DTRC.C - dump trace buffer memory.  Debugging utility for the
 *           dgdm device driver.
 *
 *  NOTE: THIS CAN BE A HUGE SECURITY HOLE IF USED IMPROPERLY AND SHOULD
 *        NOT BE SUID ROOT ON NON-DEVELOPMENT SYSTEMS.  USER BEWARE.
 *
 *****************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "rascon.h"

#define TRUE		1
#define FALSE		0
#define MAX(x,y)	(((x) > (y)) ? (x) : (y))
#define MIN(x,y)	(((x) > (y)) ? (y) : (x))

/* 8M - intentionally excessively large */
#define	TRCSZ	0x800000

#ifdef DTRC_64
#define LSEEK 	lseek64
#else
#define LSEEK 	lseek
#endif

#define MGMT_NODE	"/dev/dg/dgdm/mgmt0"

char *corefile = "/dev/kmem";
char *outputfile = "";

unsigned long trcbuf_addr;
unsigned long size;
char trcbufname[] = "trcbuf";

int raw = FALSE;
int tofile = FALSE;
int setdebug = FALSE;

unsigned char buf[TRCSZ];


void set_debug(int level);
void usage(char *progname);


main(int argc, char **argv)
{
	long	 rc;
	FILE	*out;
	int	 fd;
	int	 i;
	int	 args;
	int	 wrap;
	int	 first_pass;
        int	 debug;

	/** parse args **/

	if(argc < 2)
	  usage(argv[0]);
	
	for(i = 1; i < (argc - 1); i++) {
	  if(argv[i][0] == '-') {
		switch(argv[i][1]) {
		case 'r':
		  raw = TRUE;
		  fprintf(stderr, "Raw mode\n");
		  break;
		case 'f':
		  tofile = TRUE;
		  i++;
		  outputfile = argv[i];
		  fprintf(stderr, "Outputing to file %s\n", outputfile);
		  break;
		case 'd':
		  setdebug = TRUE;
		  i++;
                  /* get the debug level we want */
                  if((argv[i][0] == '0') && (argv[i][1] == 'x'))
                          sscanf(argv[i], "%x", &debug);
                  else
                          sscanf(argv[i], "%ul", &debug);
		  fprintf(stderr, "Setting driver debug level to 0x%x\n",
                          debug);
                  set_debug(debug);
                  /* NOT REACHED */
                  break;
		default:
		  usage(argv[0]);
		}
	  } else {
		break;
	  }
	}

	args = i;

	/* get the address we want */
	if((argv[args][0] == '0') && (argv[args][1] == 'x'))
	  sscanf(argv[args], "%x", &trcbuf_addr);
	else
	  sscanf(argv[args], "%ul", &trcbuf_addr);


	/* check if user specified size */
	if((argc - args) == 2) {
	  if((argv[args+1][0] == '0') && (argv[args+1][1] == 'x'))
		sscanf(argv[args+1], "%x", &size);
	  else
		sscanf(argv[args+1], "%ul", &size);
	} else {
	  size = 0x100000;	/* 1M default */
	}

        size = MIN(TRCSZ, size);

	fprintf(stderr, "addr=0x%lx, size=0x%lx\n", trcbuf_addr, size);
	fflush(stderr);


	/* open /dev/kmem (or other core file) */
	fd = open(corefile,O_RDONLY);
	if(fd < 0) {
		perror("/dev/kmem open failed");
		exit(1);
	}

	/* point to the trace buffer */
	rc = LSEEK(fd, trcbuf_addr, SEEK_SET);
	if (-1 == rc) {
		perror("lseek failed");
		exit(1);
	}

	/* read trace buffer */
	rc = read(fd, buf, size);
	if (-1 == rc) {
		perror("read failed");
		exit (1);
	}

	/* open the output file if we were told to */
	if(tofile) {
	  out = fopen(outputfile, "w+");
	  if(!out) {
		perror("output file open failed");
		exit (1);
	  }
	} else {
	  out = stdout;
	}


	if(!raw) {
	  /* find end of trace buffer (marked by ~) */
	  for (i=0;i<rc;i++)
		if(buf[i] == '~')
		  break;
	  wrap = i;
	  i++;

	  /* dump from tilda to end of buffer */
	  first_pass = 1;
	  for(; i < rc; i++) {
		if (buf[i] == 0) {
		  fprintf(out, "(no wrap)\n");
		  fflush(out);
		  break;
		}
		if (first_pass) {
		  if (buf[i] == 0x0a)
			first_pass = 0;
		  continue;
		}
		write(fileno(out), &buf[i], 1);
	  }

	  /* dump the rest -- begining of buffer to wrap point */
	  for (i=0;i<wrap;i++) {
		write(fileno(out), &buf[i], 1);
	  }
	} else {
	  rc = write(fileno(out), buf, size);
	  if (-1 == rc) {
		perror("write failed");
		exit (1);
	  }
	}

	  

	close(fd);
	exit(0);
}



void set_debug(int level)
{
        int	 sdebug = level;
        int	 rc = 0;
        int	 fd;
        
	fd = open(MGMT_NODE, O_RDWR);
	if(fd < 0) {
		perror("Management node open failed");
		exit(1);
	}

        rc = ioctl(fd, DIGI_SETDEBUG, &sdebug);
        if(rc < 0) {
                perror("DIGI_SETDEBUG");
        }

        exit(rc);
}





void usage(char *progname)
{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr,"\t%s trcbuf [-f outputfile] "
                "[-r] addr [size]\n", progname);
	fprintf(stderr, "\tThe -r option is for raw mode, no unwrapping "
                "will be attempted\n\n\tor\n\n");
	fprintf(stderr,"\t%s trcbuf -d debuglevel\n", progname);
	fprintf(stderr, "\tThe -d option sets the driver debug level\n\n");
	exit (1);
}

