/****************************************************************************
 * Copyright (c) 1990-2000 Digi International Inc., All Rights Reserved
 *
 * This software contains proprietary and confidential information of Digi
 * International Inc.  By accepting transfer of this copy, Recipient agrees
 * to retain this software in confidence, to prevent disclosure to others,
 * and to make no use of this software other than that for which it was
 * delivered.  This is an unpublished copyrighted work of Digi International
 * Inc.  Except as permitted by federal law, 17 USC 117, copying is strictly
 * prohibited.
 *
 * Restricted Rights Legend
 *
 * Use, duplication, or disclosure by the Government is subject to
 * restrictions set forth in sub-paragraph (c)(1)(ii) of The Rights in
 * Technical Data and Computer Software clause at DFARS 252.227-7031 or
 * subparagraphs (c)(1) and (2) of the Commercial Computer Software -
 * Restricted Rights at 48 CFR 52.227-19, as applicable.
 *
 * Digi International Inc. 11001 Bren Road East, Minnetonka, MN 55343
 *
 ****************************************************************************
 *
 * Module Name:
 *
 *  dxbuser.c
 *
 * Abstract:
 *
 *  Usermode wrapper routines for dxb streams api requests on Linux.
 *
 *  Written specifically for Tropico to provide dxb to userland for their
 *  specific needs.
 *
 * Author:
 *
 *  Jeff Randall, based upon dsuser.c by Mike Zarns and rawdxb.c
 */

#include <stdio.h>
#include <errno.h>
#include "dxbuser.h"
#include "dgdm_isdn.h"


#define RAW_DXB_DEVICE		"/dev/dg/dgdm/dxbStrm%d_%02d"
#define RAW_DXB_DEVICE_NAMELEN	128

#define OPEN_BUFFER_DATA_SIZE	512
#define MSG_BUFFER_DATA_SIZE	8192
#define DEBUG_BUFFER_SIZE	8192


static char DebugBuffer[DEBUG_BUFFER_SIZE];



/*
 * open_raw_dxb() internal routine used by dxb_open_dev()
 */
static int
open_raw_dxb(int brd)
{
    int i, fd, rc ;
    char fname[RAW_DXB_DEVICE_NAMELEN];
    
    for(i = 0; i < 100; i++) {
	snprintf(fname, RAW_DXB_DEVICE_NAMELEN, RAW_DXB_DEVICE, brd, i) ;
	if ((fd = open(fname, O_RDWR)) >= 0) {
	    rc = ioctl(fd, I_VERYRAW, 0);
	    if(rc < 0) {
		perror("ioctl") ;
		close(fd) ;
		return(-1);
	    }
	    return(fd);
	}
    }
    
    return(-1);
}



/*
 * dxb_open_dev()
 *
 * This is how to send the open message.
 *
 * The original ISDN implementation kidnapped the data portion of the
 * open message to indicate receive buffer size.  Since the data
 * portion is actually used in some cases, the driver was modified
 * to take a 4-byte interger off of the begining of the data section, and
 * pass the rest to the DXB open.
 *
 */
int
dxb_open_dev(int brd, char *devnm, int maxsize, int datasz, char *data)
{
    int fd ;
    char buffer[OPEN_BUFFER_DATA_SIZE] ;
    DXB_STRM_MESSAGE *open_msg = (DXB_STRM_MESSAGE *) buffer ;
    
    if( (fd = open_raw_dxb(brd)) == -1 ) {
	perror("open_raw_device") ;
	return -1 ;
    }

    open_msg->command = DXB_TYPE_OPEN ;
    open_msg->priority = DXB_PRI_NORMAL ;
 
    open_msg->control.maxlen = maxsize ;
    open_msg->control.len = strlen(devnm)+1 ;
    open_msg->control.len = (open_msg->control.len+3) & ~0x03 ;
    open_msg->control.buffer_offset = sizeof(DXB_STRM_MESSAGE) ;

    open_msg->data.maxlen = sizeof(int) + datasz ;
    open_msg->data.len = sizeof(int) + datasz ;
    open_msg->data.buffer_offset = (open_msg->control.buffer_offset +
				    open_msg->control.len);
    memcpy(&buffer[open_msg->control.buffer_offset], devnm, strlen(devnm)+1) ;
    *((int *) &buffer[open_msg->data.buffer_offset]) = maxsize ;
    if(datasz)
	memcpy(&buffer[open_msg->data.buffer_offset + 4], data, datasz) ;
    
    write(fd, buffer, OPEN_BUFFER_DATA_SIZE) ;
    return(fd);
}



/*
 * dxb_close_dev()
 *
 * Close the DXB stream.
 */
int
dxb_close_dev(int fd)
{
    int rc;

    rc = close(fd);
    if(rc < 0)
	perror("close");

    return(rc);
}



/*
 * dxb_putmsg()
 *
 * This will send a dxb message to the channel.
 */
int
dxb_putmsg(int fd, struct strbuf *cbuf, struct strbuf *dbuf, int flags)
{
    char buffer[MSG_BUFFER_DATA_SIZE + sizeof(DXB_STRM_MESSAGE)] ;
    DXB_STRM_MESSAGE *msg = (DXB_STRM_MESSAGE *) buffer ;
    int n ;
 
    if (!cbuf && !dbuf) return 0 ;
 
    if ((cbuf && !cbuf->buf) || (dbuf && !dbuf->buf)) {
	return(-ENOMEM);
    }

    msg->command = (cbuf ? DXB_TYPE_CDATA : DXB_TYPE_DATA) ;
    msg->priority = DXB_PRI_NORMAL ;
    if(flags)
	msg->priority = DXB_PRI_HIGH ;
    msg->control.buffer_offset = sizeof(DXB_STRM_MESSAGE) ;

    if (cbuf) {
	msg->control.len = cbuf->len ;
	memcpy(&buffer[msg->control.buffer_offset],
	       cbuf->buf,
	       cbuf->len) ;
    } else {
	msg->control.len = 0 ;
    }

    if (dbuf) {
	msg->data.buffer_offset = (msg->control.buffer_offset +
				   msg->control.len);
	msg->data.buffer_offset = (msg->data.buffer_offset + 3) & ~0x03 ;
	msg->data.len = dbuf->len ;
	memcpy(&buffer[msg->data.buffer_offset], dbuf->buf, dbuf->len) ;
    } else {
	msg->data.len = 0 ;
    }
 
    n = write(fd, &buffer[0], (sizeof(DXB_STRM_MESSAGE) +
			       msg->control.len+msg->data.len));
 
    if (n < 0) {
	printf("write returned %x on %x bytes, errno = %x\n", n,
	       sizeof(DXB_STRM_MESSAGE) + msg->control.len + msg->data.len,
	       errno) ;
    }

    return(n);
}



/*
 * dxb_getmsg()
 *
 * Get a DXB message from the channel
 *
 */
int
dxb_getmsg(int fd, struct strbuf *cbuf, struct strbuf *dbuf, int *flags)
{
    char buffer[MSG_BUFFER_DATA_SIZE + sizeof(DXB_STRM_MESSAGE)] ;
    DXB_STRM_MESSAGE *msg = (DXB_STRM_MESSAGE *) buffer ;
    int n ;

    if (!cbuf && !dbuf) {
        fprintf(stderr, "no cbuf or dbuf\n") ;
        return(-ENOSPC);
    }

    if ((cbuf && !cbuf->buf) || (dbuf && !dbuf->buf)) {
        fprintf(stderr, "cbuf or dbuf missing pointer\n") ;
        return(-ENXIO);
    }

    msg->control.len = msg->data.len = 0 ;

    /*
     * *INTERESTING* IMPLEMENTATION NOTE:
     *
     * The read below will first read the contents of the
     * user space buffer to find the maximum lengths; then it
     * coppies the result back into the buffer.  So, you can't
     * simply do a regular read with this driver!
     */

    if (cbuf)
        msg->control.maxlen = cbuf->maxlen ;
    else
        msg->control.maxlen = 0 ;
 
    if (dbuf)
        msg->data.maxlen = dbuf->maxlen ;
    else
        msg->data.maxlen = 0 ;

    msg->command = (cbuf ? DXB_TYPE_CDATA : DXB_TYPE_DATA) ;
    msg->priority = DXB_PRI_NORMAL ;
    msg->control.buffer_offset = sizeof(DXB_STRM_MESSAGE) ;

    if ((n = read(fd, &buffer[0], msg->control.maxlen + 
		  msg->data.maxlen + sizeof(DXB_STRM_MESSAGE))) < 0)
        return(n);

    if (n < (msg->control.len + msg->data.len)) {
        fprintf(stderr, "error: %d bytes of payload, control len(%d) data "
		"len(%d)\n", n, msg->control.len, msg->data.len) ;
        return(-ENOMEM);
    }

    if (cbuf) {
        cbuf->len = msg->control.len ;
        memcpy(cbuf->buf, &buffer[sizeof(DXB_STRM_MESSAGE)], cbuf->len) ;
    }

    if (dbuf) {
        dbuf->len = msg->data.len ;
        memcpy(dbuf->buf, &buffer[sizeof(DXB_STRM_MESSAGE) +
				 msg->control.maxlen],
	       dbuf->len) ;
    }

    /* Temporary patch for empty control blocks */
    if (cbuf && !cbuf->len)
        cbuf->len = -1 ;
 
    return(0);
}



/*
 * dxb_ioctl()
 *
 * IOCTL requests to the dxb stream.
 *
 */
int
dxb_ioctl(int fd, struct strioctl *sioctl)
{
    int rc;

    rc = ioctl(fd, I_STR, sioctl);

    if(rc == -1)
	perror("errno");

    return(rc);
}



/*
 * dxb_flush()
 *
 * Requests that the dxb queue be purged.
 *
 */
int
dxb_flush(int fd)
{
    char buffer[MSG_BUFFER_DATA_SIZE + sizeof(DXB_STRM_MESSAGE)] ;
    DXB_STRM_MESSAGE *msg = (DXB_STRM_MESSAGE *) buffer ;
    int n ;
 
    msg->command = DXB_TYPE_FLUSH ;
    msg->priority = DXB_PRI_HIGH ;
    msg->control.buffer_offset = sizeof(DXB_STRM_MESSAGE) ;
    msg->control.len = 0 ;
    msg->data.len = 0 ;

    n = write(fd, &buffer[0], sizeof(DXB_STRM_MESSAGE)) ;
    if (n < 0) {
	printf("Flush returned %x, errno = %x\n", n, errno) ;
    }
    return(n);
}




